import re
import paramiko
from subprocess import Popen, PIPE, STDOUT
from CommandContainers.PbsCommands import PbsCommands


class Pbs:
    DEPENDENCY_AFTER_OK = 'afterok'
    DEPENDENCY_AFTER_ANY = 'afterany'

    def __init__(self, *commands):
        self.pbs_commands = PbsCommands()
        self.commands = commands
        self.submitted = False
        self.dependency = None
        self.id = None

    def get_string(self):
        parts = []

        for command in (self.pbs_commands,) + self.commands:
            string = command.get_header() + '\n' + command.__str__()
            parts.append(string)

        return '\n\n'.join(parts)

    def depends(self, job, type=DEPENDENCY_AFTER_OK):
        if not isinstance(job, Pbs):
            raise Exception('Given argument should be an instance of Pbs')

        self.dependency = job
        self.dependency_type = type
        return self

    def run_command(self, command, pipe_in):
        p = Popen(command, stdout=PIPE, stdin=PIPE, stderr=STDOUT)
        stdout = p.communicate(input=pipe_in)[0]
        p.stdin.close()

        return stdout

    def submit(self, program='qsub'):
        input = self.get_string()

        if isinstance(self.dependency, Pbs):
            if not self.dependency.submitted:
                self.dependency.submit()

            command = [program, '-W', 'depend=%s:%s' % (self.dependency_type, self.dependency.id)]
        else:
            command = program

        stdout = self.run_command(command, input)

        self.id = re.search('(\d+)', stdout).group(1)
        self.submitted = True
        return self

    def __getattr__(self, name):
        def function(*args):
            if name in dir(self.pbs_commands):
                method = getattr(self.pbs_commands, name)
                if callable(method):
                    method(*args)
                    return self

        return function

    def get_name(self):
        return self.pbs_commands._name

    def __str__(self):
        return self.get_string()


class SshPbs(Pbs):
    def __init__(self, *commands, **kwargs):
        assert 'ssh' in kwargs
        assert isinstance(kwargs.get('ssh'), paramiko.SSHClient)
        Pbs.__init__(self, *commands)
        self.ssh = kwargs.get('ssh')

    def run_command(self, command, pipe_in):
        stdin, stdout, stderr = self.ssh.exec_command(command)
        stdin.write(input)
        stdin.flush()
        stdin.channel.shutdown_write()
        return stdout.read()



class CommandContainerInterface:
    def __str__(self):
        raise NotImplementedError("A command container should implement __str__ method")

    def get_name(self):
        raise NotImplementedError("A command container should implement get_name method")

    def get_header(self):
        return '# %s' % self.get_name()



from CommandContainerInterface import CommandContainerInterface


class ShellCommands(CommandContainerInterface):
    def __init__(self, name='Shell Commands'):
        self.commands = []
        self._name = name

    def __getattr__(self, name):
        def function(*args):
            self.commands.append('%s %s' % (name, ' '.join(args)))
            return self

        return function

    def append(self, command):
        self.commands.append(command)
        return self

    def get_name(self):
        return self._name

    def __str__(self):
        return '\n'.join(self.commands)



from CommandContainerInterface import CommandContainerInterface
import re


class PbsCommands(CommandContainerInterface):
    TYPE_SERIAL = 0
    TYPE_OPEN_MP = 1
    TYPE_MPI = 2
    TYPE_HYBRID = 3

    def __init__(self, name='Portable Batch System Commands'):
        self.command_container_name = name
        self.commands = []

    def get_name(self):
        return self.command_container_name

    def serial(self):
        self.type = PbsCommands.TYPE_SERIAL
        return self

    def open_mp(self, processors):
        self.type = PbsCommands.TYPE_OPEN_MP
        self.processors_per_node = processors
        return self

    def mpi(self, nodes):
        self.type = PbsCommands.TYPE_MPI
        self.nodes = nodes
        return self

    def hybrid(self, nodes, processors_per_node):
        self.type = PbsCommands.TYPE_HYBRID
        self.nodes = nodes
        self.processors_per_node = processors_per_node
        return self

    def queue(self, queue):
        self._queue = queue
        return self

    def accounting_group(self, accounting_group):
        self._accounting_group = accounting_group
        return self

    def name(self, name):
        self._name = ''.join(ch for ch in name if ch.isalnum())[:15]
        return self

    def memory_per_process(self, memory_per_process):
        self._memory_per_process = memory_per_process

    def wall_time(self, time):
        self._wall_time = time
        return self

    def cpu_time(self, time):
        self._cpu_time = time
        return self

    def job_array_size(self, size):
        self._job_array_size = size

    def add_pbs_command(self, string):
        self.commands.append(string)

    def add_attribute(self, key, value):
        self.commands.append('#PBS -l %s=%s' % (key, value))

    def __str__(self):
        lines = []

        # Prepare the part of PBS Commands for processors and nodes
        if hasattr(self, 'type'):
            if self.type == PbsCommands.TYPE_SERIAL:
                lines.append('#PBS -l nodes=1:ppn=1')
                lines.append('#PBS -l procs=1')
            elif self.type == PbsCommands.TYPE_OPEN_MP:
                lines.append('#PBS -l nodes=1:ppn=%d' % self.processors_per_node)
            elif self.type == PbsCommands.TYPE_MPI:
                lines.append('#PBS -l nodes=%d:ppn=1' % self.nodes)
                lines.append('#PBS -l procs=%d' % self.nodes)
            elif self.type == PbsCommands.TYPE_HYBRID:
                lines.append('#PBS -l nodes=%d:ppn=%d' % (self.nodes, self.processors_per_node))

        if hasattr(self, '_name'):
            lines.append('#PBS -N %s' % self._name)

        if hasattr(self, '_accounting_group'):
            lines.append('#PBS -A %s' % self._accounting_group)

        if hasattr(self, '_queue'):
            lines.append('#PBS -q %s' % self._queue)

        if hasattr(self, '_wall_time'):
            lines.append('#PBS -l walltime=%s' % format_time_limit(self._wall_time))

            if not hasattr(self, '_cpu_time'):
                time = None

                if hasattr(self, 'type'):
                    if self.type == PbsCommands.TYPE_SERIAL:
                        time = self._wall_time
                    elif self.type == PbsCommands.TYPE_OPEN_MP:
                        time = self.processors_per_node * timespec_to_seconds(self._wall_time)
                    elif self.type == PbsCommands.TYPE_MPI:
                        time = self.nodes * timespec_to_seconds(self._wall_time)
                    elif self.type == PbsCommands.TYPE_HYBRID:
                        time = self.nodes * self.processors_per_node * timespec_to_seconds(self._wall_time)

                if time is not None:
                    lines.append('#PBS -l cput=%s' % format_time_limit(time))

        if hasattr(self, '_cpu_time'):
            lines.append('#PBS -l cput=%s' % format_time_limit(self._cpu_time))

        if hasattr(self, '_memory_per_process'):
            lines.append('#PBS -l pvmem=%s' % self._memory_per_process)

        if hasattr(self, '_job_array_size'):
            lines.append('#PBS -t %s' % self._job_array_size)

        return '\n'.join(lines)


def timespec_to_seconds(string):
    string = str(string)

    pattern = '^(\d+)([dhm]?)$'
    coefficients = {'d': 86400, 'h': 3600, 'm': 60, '': 3600}

    results = re.search(pattern, string)

    if results is not None:
        digits = int(results.group(1))
        unit = results.group(2)

        return digits * coefficients[unit]
    else:
        raise Exception('The given time "%s" is invalid.' % string)


def format_time_limit(input):
    if isinstance(input, str):
        seconds = timespec_to_seconds(input)
    else:
        seconds = int(input)

    m, s = divmod(seconds, 60)
    h, m = divmod(m, 60)
    return '%02d:%02d:%02d' % (h, m, s)



#!/usr/bin/env python
import click
import re
import paramiko
import os
import glob
from tabulate import tabulate
from pbs_generator import SshPbs, ShellCommands, PbsCommands

ip_address_regex = '(([0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5])\.){3}([0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5])'
host_name_regex = '[a-zA-Z0-9](?:(?:[a-zA-Z0-9-]*|(?<!-)\.(?![-.]))*[a-zA-Z0-9]+)?'
username_regex = '[a-z_][a-z0-9_-]*[$]?'
port_regex = '[0-9]{1,4}|[1-5][0-9]{4}|6[0-4][0-9]{3}|65[0-4][0-9]{2}|655[0-2][0-9]|6553[0-5]'
ssh_url_regex = '(%s)\@(%s|%s)(\:(%s))?' % (username_regex, ip_address_regex, host_name_regex, port_regex)


def parse_ssh_url(url):
    search = re.search(ssh_url_regex, url)
    assert search is not None, 'Entered ssh url is invalid: %s' % url
    username = search.group(1)
    hostname = search.group(2)
    port = int(search.group(7)) if search.group(7) is not None else 22

    return [username, hostname, port]


def get_connection(url, prompt_password=False):
    ssh_known_hosts = os.path.expanduser('~/.ssh/known_hosts')
    ssh_public_key = os.path.expanduser('~/.ssh/id_rsa.pub')

    ssh = paramiko.SSHClient()
    ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
    ssh.get_host_keys().load(ssh_known_hosts)
    username, hostname, port = parse_ssh_url(url)

    args = {
        'username': str(username),
        'port': port,
    }

    if prompt_password:
        args['password'] = click.prompt('Enter password for %s@%s' % (username, hostname), hide_input=True)
    else:
        args['key_filename'] = ssh_public_key

    ssh.connect(hostname, **args)
    return ssh


@click.command()
@click.option('-p', is_flag=True, default=False, help='Either prompt you for password or use id_rsa')
@click.option('-d', type=click.Path(), default='~', help='The working directory on remote machine')
@click.option('--files', type=click.Path(),
              help='The files that you want to work with as input files. Use local paths on your machine. '
                   'The files will be automatically copied to the server\n'
                   'You can use wildcard path like:\n --files="inputs/*.inp"\n'
                   'It will automatically loop over all the files and run the command for each file. '
                   'The environment variables `$INPUT_FILE` and `$INPUT_NAME` will be made '
                   'that you can use inside your command'
              )
@click.option('--serial', is_flag=True, help='Requests 1 processor on 1 node.')
@click.option('--openmp', type=click.INT, nargs=1,
              help='This should be used for jobs that use shared memory parallelization '
                   '(e.g., OpenMP) or when communication across an interconnect severely '
                   'impacts job performance.')
@click.option('--mpi', type=click.INT, nargs=1, help='This mode of parallelization is suitable for '
                                                     'Message Passing Interface (MPI) jobs.')
@click.option('--hybrid', type=click.INT, nargs=2, help='Requests for N nodes with M processors.')
@click.option('--attribute', '-a', help='Request node attribute eg. submit -a switch2:xeon')
@click.option('--mpp', help='Memory per process')
@click.option('--queue', '-q', help='Specify a queue to run on')
@click.option('--time', '-t', type=click.STRING, help='Wall time', default='24')
@click.option('--debug', '-D', is_flag=True, help='Don\'t submit job, print debugging info')
@click.option('--job-name', '-j', help='Job name')
@click.argument('ssh_url', type=click.STRING)
@click.argument('command', type=click.STRING, required=True)
def ssh(p, d, files, serial, openmp, mpi, hybrid, mpp, attribute, queue, time, debug, job_name, ssh_url, command):
    ssh = get_connection(ssh_url, p)
    sftp = ssh.open_sftp()

    commands = ShellCommands().cd(d).append(command)
    pbs_command = get_pbs_command(serial, openmp, mpi, hybrid, mpp, attribute, queue, time)

    if files is not None:
        files_array = glob.glob(files)
        jobs = []

        with click.progressbar(files_array, label='Uploading the files' if not debug else 'Printing PBS') as bar:
            for file in bar:
                file_name = os.path.basename(file)
                name, extension = os.path.splitext(file_name)
                remote_file_path = os.path.join(d, file_name)

                export_commands = ShellCommands() \
                    .append('export INPUT_FILE="%s"' % remote_file_path) \
                    .append('export INPUT_NAME="%s"' % name)

                job = SshPbs(pbs_command, export_commands, commands, ssh=ssh).name(name)

                if debug:
                    click.echo('\n\nJob "%s":' % name)
                    click.echo('=' * 58)
                    click.echo(job.get_string())
                    click.echo('=' * 58)
                else:
                    sftp.put(file, remote_file_path)
                    job.submit()
                    job.input_file = file_name
                    jobs.append(job)

        if not debug:
            click.echo('\n')
            table = map(lambda job: [job.input_file, job.get_name(), job.id], jobs)
            click.echo(tabulate(table, headers=['Input file', 'Job name', 'Job ID'], tablefmt="rst"))

    ssh.close()


def get_pbs_command(serial, openmp, mpi, hybrid, mpp, attribute, queue, time):
    pbs_commands = PbsCommands('Common PBS commands from CLI options')

    if openmp:
        pbs_commands.open_mp(openmp)
    elif mpi:
        pbs_commands.mpi(mpi)
    elif hybrid:
        pbs_commands.hybrid(*hybrid)
    elif serial:
        pbs_commands.serial()

    if mpp is not None:
        pbs_commands.memory_per_process(mpp)

    if attribute is not None:
        pbs_commands.add_pbs_command('#PBS -l %s' % attribute)

    if queue is not None:
        pbs_commands.queue(queue).accounting_group(queue)

    if time is not None:
        pbs_commands.wall_time(time)

    return pbs_commands


if __name__ == '__main__':
    ssh()



from CommandContainerInterface import CommandContainerInterface


class Modules(CommandContainerInterface):
    def __init__(self):
        self._purge = False
        self.loaded_modules = []
        self.unloaded_modules = []

    def purge(self):
        self._purge = True
        return self

    def get_name(self):
        return 'Modules'

    def load(self, module):
        if isinstance(module, list):
            self.loaded_modules = self.loaded_modules + module
        else:
            self.loaded_modules.append(module)

        return self

    def unload(self, module):
        if isinstance(module, list):
            self.unloaded_modules = self.unloaded_modules + module
        else:
            self.unloaded_modules.append(module)

        return self

    def __str__(self):
        lines = []

        if self._purge:
            lines.append('module purge')

        lines = lines + map(lambda module: 'module unload %s' % module, self.unloaded_modules)
        lines = lines + map(lambda module: 'module load %s' % module, self.loaded_modules)

        return '\n'.join(lines)



from Pbs import Pbs
from Pbs import SshPbs
from CommandContainers.Modules import Modules
from CommandContainers.ShellCommands import ShellCommands
from CommandContainers.PbsCommands import PbsCommands



from setuptools import setup, find_packages

setup(
    name='pbs-generator',
    packages=find_packages(),
    version='1.3',
    entry_points = {
        'console_scripts': ['nicesub=pbs_generator.cli:ssh'],
    },
    install_requires=['Click', 'tabulate', 'paramiko'],
    description = 'A PBS script generator',
    author = 'Mohamad Mohebifar',
    author_email = 'mmohebifar@mun.ca',
    url = 'https://github.com/RowleyGroup/pbs-generator',
    download_url = 'https://github.com/RowleyGroup/pbs-generator/tarball/1.0',
    keywords = ['pbs-generator', 'rowley'],
    classifiers = [],
)
